/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.io.File;
import java.util.List;
import java.util.ArrayList;

/**

*/
public class InteractiveModels_R2 implements KeyListener, ComponentListener
{
   private final FrameBufferPanel fbp; // The event handlers need
   private final Scene scene;          // access to these fields.
   private final List<Model> modelArray = new ArrayList<>();
   private int currentModel = 1;
   private double[] deltaX = null;
   private double[] deltaY = null;
   private double[] deltaZ = null;
   private double[] deltaS = null;
   private boolean displayTransformations = false;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveModels_R2()
   {
      // Create the Scene object that we shall render
      scene = new Scene();

      // Create several Model objects.
      modelArray.add( new ObjSimpleModel(new File("assets/apple.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cow.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/galleon.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/teapot.obj")) );
      modelArray.add( new ObjSimpleModel(new File("assets/cessna.obj")) );
      modelArray.add( new Sphere(1.0, 30, 30) );
      modelArray.add( new Cylinder(0.5, 1.0, 20, 20) );
      modelArray.add( new Torus(0.75, 0.25, 25, 25) );
      modelArray.add( new Cube2(15, 15, 15) );
      modelArray.add( new ObjSimpleModel(new File("assets/small_rhombicosidodecahedron.obj")) );
      final Model floor = new PanelXZ(-2, 2, -2, 0);  // floor
      final Model wall  = new PanelXY(-4, 4, -1, 3);  // wall
      final Model airplane = new ObjSimpleModel(new File("assets/cessna.obj"));

      // Add four models to the Scene.
      scene.addModel(modelArray.get(1), // cow
                     floor,
                     wall,
                     airplane);

      deltaX = new double[modelArray.size()];
      deltaY = new double[modelArray.size()];
      deltaZ = new double[modelArray.size()];
      deltaS = new double[modelArray.size()];
      java.util.Arrays.fill(deltaS, 1.0);

      // Position the wall, floor and airplane.
      for (int i = 0; i < wall.vertexList.size(); ++i)
      {
         final Vertex v = wall.vertexList.get(i);
         wall.vertexList.set(i, new Vertex(v.x,
                                           v.y,
                                           v.z - 4));
      }
      for (int i = 0; i < floor.vertexList.size(); ++i)
      {
         final Vertex v = floor.vertexList.get(i);
         floor.vertexList.set(i, new Vertex(v.x,
                                            v.y - 1,
                                            v.z - 2));
      }
      for (int i = 0; i < airplane.vertexList.size(); ++i)
      {
         final Vertex v = airplane.vertexList.get(i);
         airplane.vertexList.set(i, new Vertex(v.x + 2.5,
                                               v.y,
                                               v.z - 2));
      }
      // Push the models away from where the camera is.
      for (final Model m : modelArray)
      {
         for (int i = 0; i < m.vertexList.size(); ++i)
         {
            final Vertex v = m.vertexList.get(i);
            m.vertexList.set(i, new Vertex(v.x,
                                           v.y,
                                           v.z - 2));
         }
      }

      Rasterize_Clip.doClipping = true;


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that holds a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         modelArray.get(currentModel).debug = ! modelArray.get(currentModel).debug;
      }
      else if ('D' == c)
      {
         Rasterize_Clip.debug = ! Rasterize_Clip.debug;
      }
      else if ('c' == c)
      {
         Rasterize_Clip.doClipping = ! Rasterize_Clip.doClipping;
         System.out.print("Clipping is turned ");
         System.out.println(Rasterize_Clip.doClipping ? "On" : "Off");
      }
      else if ('/' == c)
      {
         currentModel = (currentModel + 1) % modelArray.size();
         scene.setModel( 0, modelArray.get(currentModel) );
      }
      else if ('?' == c)
      {
         currentModel = (currentModel - 1);
         if (currentModel < 0) currentModel = modelArray.size() - 1;
         scene.setModel( 0, modelArray.get(currentModel) );
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
         final String p = scene.camera.perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('m' == c) //display transformation information
      {
         displayTransformations = ! displayTransformations;
      }
      else if ('s' == c) // Scale the current model 10% smaller.
      {
         // This does NOT work with perspective projection.
         scaleModel(currentModel, 1.0/1.1);
      }
      else if ('S' == c) // Scale the current model 10% larger.
      {
         // This does NOT work with perspective projection.
         scaleModel(currentModel, 1.1);
      }
      else if ('x' == c) // Translate the current model.
      {
         moveModel(currentModel, -0.1, 0, 0); // left
      }
      else if ('X' == c)
      {
         moveModel(currentModel, +0.1, 0, 0); // right
      }
      else if ('y' == c)
      {
         moveModel(currentModel, 0, -0.1, 0); // down
      }
      else if ('Y' == c)
      {
         moveModel(currentModel, 0, +0.1, 0); // up
      }
      else if ('z' == c)
      {
         moveModel(currentModel, 0, 0, -0.1); // back
      }
      else if ('Z' == c)
      {
         moveModel(currentModel, 0, 0, +0.1); // forward
      }

      if (displayTransformations &&
            ('m'==c||'/'==c||'?'==c
           ||'s'==c||'x'==c||'y'==c||'z'==c
           ||'S'==c||'X'==c||'Y'==c||'Z'==c))
      {
         System.out.printf("deltaX = %.2f, deltaY = %.2f, " +
                           "deltaZ = %.2f, deltaS = %.2f\n",
                           deltaX[currentModel], deltaY[currentModel],
                           deltaZ[currentModel], deltaS[currentModel]);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      fbp.update();
   }


   private void scaleModel(int currentModel, double deltaS)
   {
      this.deltaS[currentModel] *= deltaS;
      final Model m = scene.getModel(0);
      for (int i = 0; i < m.vertexList.size(); ++i)
      {
         final Vertex v = m.vertexList.get(i);
         m.vertexList.set(i, new Vertex(v.x * deltaS,
                                        v.y * deltaS,
                                        v.z * deltaS));
      }
   }


   private void moveModel(int currentModel, double deltaX, double deltaY, double deltaZ)
   {
      this.deltaX[currentModel] += deltaX;
      this.deltaY[currentModel] += deltaY;
      this.deltaZ[currentModel] += deltaZ;
      final Model m = scene.getModel(0);
      for (int i = 0; i < m.vertexList.size(); ++i)
      {
         final Vertex v = m.vertexList.get(i);
         m.vertexList.set(i, new Vertex(v.x + deltaX,
                                        v.y + deltaY,
                                        v.z + deltaZ));
      }
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        fbp.getTopLevelAncestor().getWidth(),
                        fbp.getTopLevelAncestor().getHeight(),
                        fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveModels_R2()
      );
   }


   private static void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off for the current model.");
      System.out.println("Use the '/' key to cycle through the models.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the current model along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the current model.");
      System.out.println("Use the 'm' key to toggle the display of transformation information.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
