/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
<pre>{@code
                        |
                      2 + v[1]
                        |
                        |
                        |
                        |
                        |
                  |     |
                2 +     |
                  |     |
                  |     |
                  |   1 +
            y     |     |
            |     |     |        -z
          2 +     |     |       /
            |     |     |      /
            |     |     |     /
            |     |     |    /       This 3D space is called "camera space".
            |   1 +     |   /
            |     |     |  /
            |     |     | /
            |     |     |/
            |     |  -2 +-----------------+-----------------+---- y = -2 plane
            |     |    /v[2]              1                 v[0]
          1 +     |   /
            |     |  /
            |     | /
            |     |/
            |  -1 +-----------------+----- image plane (y = -1)
            |    /                  1
            |   /
            |  /
            | /
            |/
     Camera +-----------------+---------> x-axis
           /0                 1
          /
         z-axis
}</pre>
   Render a wireframe triangle. This is just about the
   simplest possible model. It is useful for debugging.
*/
public class InteractiveTriangle_R2 implements KeyListener, ComponentListener
{
   private final FrameBufferPanel fbp; // The event handlers need
   private final Scene scene;          // access to these fields.
   private double deltaX = 0.0;
   private double deltaY = 0.0;
   private double deltaZ = 0.0;
   private double deltaS = 1.0;
   private boolean displayTransformations = false;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveTriangle_R2()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Create a Model object to hold the geometry.
      final Model model = new Model("triangle");

      // Add the Model to the Scene.
      scene.addModel(model);

      // Create the geometry for the Model.
      // Vertices.
      model.addVertex(new Vertex(2.0, 0.0, -2.0),
                      new Vertex(0.0, 2.0, -2.0),
                      new Vertex(0.0, 0.0, -2.0));
      // Lines segments.
      model.addLineSegment(new LineSegment(0, 1),
                           new LineSegment(1, 2),
                           new LineSegment(2, 0));

      Rasterize_Clip.doClipping = false; // turn off clipping
      System.out.println(model);


      // Define initial dimensions for a FrameBuffer.
      final int width  = 512;
      final int height = 512;

      // Create a FrameBufferPanel that holds a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
         Rasterize_Clip.debug = Pipeline.debug;
      }
      else if ('D' == c)
      {
         Rasterize_Clip.debug = ! Rasterize_Clip.debug;
      }
      else if ('c' == c)
      {
         Rasterize_Clip.doClipping = ! Rasterize_Clip.doClipping;
         System.out.print("Clipping is turned ");
         System.out.println(Rasterize_Clip.doClipping ? "On" : "Off");
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
         final String p = scene.camera.perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('m' == c) //display transformation information
      {
         displayTransformations = ! displayTransformations;
      }
      else if ('s' == c) // Scale the triangle 10% smaller.
      {
         scaleModel(1.0/1.1); // does NOT work with perspective projection
      }
      else if ('S' == c) // Scale the triangle 10% larger.
      {
         scaleModel(1.1); // does NOT work with perspective projection
      }
      else if ('x' == c)
      {
         moveModel(-0.1, 0, 0); // left
      }
      else if ('X' == c)
      {
         moveModel(+0.1, 0, 0); // right
      }
      else if ('y' == c)
      {
         moveModel(0, -0.1, 0); // down
      }
      else if ('Y' == c)
      {
         moveModel(0, +0.1, 0);  // up
      }
      else if ('z' == c)
      {
         moveModel(0, 0, -0.1);  // back
      }
      else if ('Z' == c)
      {
         moveModel(0, 0, +0.1); // forward
      }

      if (displayTransformations && ('m'==c
           ||'s'==c||'x'==c||'y'==c||'z'==c
           ||'S'==c||'X'==c||'Y'==c||'Z'==c))
      {
         System.out.printf("deltaX = %.2f, deltaY = %.2f, " +
                           "deltaZ = %.2f, deltaS = %.2f\n",
                            deltaX,        deltaY,
                            deltaZ,        deltaS);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      fbp.update();
   }


   private void scaleModel(double deltaS)
   {
      this.deltaS *= deltaS;
      final Model m = scene.getModel(0);
      for (int i = 0; i < m.vertexList.size(); ++i)
      {
         final Vertex v = m.vertexList.get(i);
         m.vertexList.set(i, new Vertex(v.x * deltaS,
                                        v.y * deltaS,
                                        v.z * deltaS));
      }
   }


   private void moveModel(double deltaX, double deltaY, double deltaZ)
   {
      this.deltaX += deltaX;
      this.deltaY += deltaY;
      this.deltaZ += deltaZ;
      final Model m = scene.getModel(0);
      for (int i = 0; i < m.vertexList.size(); ++i)
      {
         final Vertex v = m.vertexList.get(i);
         m.vertexList.set(i, new Vertex(v.x + deltaX,
                                        v.y + deltaY,
                                        v.z + deltaZ));
      }
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        fbp.getTopLevelAncestor().getWidth(),
                        fbp.getTopLevelAncestor().getHeight(),
                        fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveTriangle_R2()
      );
   }


   private static void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the triangle along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the triangle.");
      System.out.println("Use the 'm' key to toggle the display of transformation information.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
